package util;

import java.util.Iterator;
import java.util.NoSuchElementException;
/**
 * Doppelt verkettete zirkuläre Liste.
 */
public class LinkedObjects implements Iterable<Object> {
    /*
     * Invariante:
     * head zeigt auf dem Kopfknoten der Liste
     * head.next zeigt auf den ersten Listenknoten
     * head.prev zeigt auf den letzten Listenknoten
     * head.obj = null
     * fuer jeden Listenknoten k zeigt
     *     k.obj auf das gespeicherte Objekt
     *     k.prev auf den Vorgaenger
     *     k.next auf den Nachfolger
     * Vor dem ersten Knoten ist head und nach dem letzten Knoten ist head.
     * In einer leeren Liste zeigt der Head-Knoten auf sich selbst.
     * 
     */
    private static class Link {
        final Object obj;
        Link next;
        Link prev;

        /**
         * Wird nur bei der Initialisierung aufgerufen!
         */
        Link() {
            obj = null;
            this.next = this.prev = this;
        }

        /**
         * Erzeugt einen neuen Knoten.
         * @param obj   gespeicherter Inhalt
         * @param prev  Vorgaenger
         * @param next  Nachfolger
         */
        Link(Object obj, Link prev, Link next) {
            this.obj = obj;
            this.prev = prev;
            this.next = next;
        }
    }
    private final Link head = new Link();
    
    /**
     * Haengt einen Wert am Ende an.
     * @param obj neuer Wert
     */
    public void add(Object obj) {
        Link neu = new Link(obj, head.prev, head);
        head.prev = head.prev.next = neu;
    }
    
    /**
     * Fuegt einen Wert am Anfang ein.
     * @param obj neuer Wert
     */
    public void addFirst(Object obj) {
        Link neu = new Link(obj, head, head.next);
        head.next = head.next.prev = neu;
    }
    
    /**
     * Entfernt den ersten Wert.
     * @return entfernter Wert
     * @throws NoSuchElementException wenn die Liste leer ist.
     */
    public Object removeFirst() {
        if (isEmpty()) throw new NoSuchElementException();
        Link toRemove = head.next;
        head.next = toRemove.next;
        head.next.prev = head;
        return toRemove.obj;
    }
    
    /**
     * Entfernt den letzten Wert.
     * @return entfernter Wert
     * @throws NoSuchElementException wenn die Liste leer ist.
     */
    public Object removeLast() {
        if (isEmpty()) throw new NoSuchElementException();
        Link toRemove = head.prev;
        head.prev = toRemove.prev;
        head.prev.next = head;
        return toRemove.obj;
    }

    /**
     * Pruef ob die Liste leer ist.
     * @return {@code true} wenn die Liste leer ist
     */
    public boolean isEmpty() {
        return head.next == head;
    }

    /**
     * Gibt ein Iterator-Objekt zurueck. Man kann damit die Liste mit einer
     * foreach-Schleife durchlaufen.
     */
    @Override
    public Iterator<Object> iterator() {
        return new Iterator<Object>() {
            Link current = head.next;

            @Override
            public boolean hasNext() {
                return current != head;
            }
            @Override
            public Object next() {
                if (!hasNext()) throw new NoSuchElementException();
                Object result = current.obj;
                current = current.next;
                return result;
            }
            @Override
            public void remove() {
                Link toRemove = current.prev;
                if (toRemove == head) throw new NoSuchElementException();
                current.prev = toRemove.prev;
                current.prev.next = current;
            }
        };
    }
}
